/**
 * This work is provided under the terms of the CREATIVE COMMONS PUBLIC
 * LICENSE. This work is protected by copyright and/or other applicable
 * law. Any use of the work other than as authorized under this license
 * or copyright law is prohibited.
 *
 * http://creativecommons.org/licenses/by-nc-sa/2.5/
 *
 * Copyright (C) 2016 OX Software GmbH
 * Mail: info@open-xchange.com
 *
 * @author Daniel Rentz <daniel.rentz@open-xchange.com>
 */

define('io.ox/office/baseframework/view/toolbar', [
    'io.ox/office/tk/utils',
    'io.ox/office/tk/forms',
    'io.ox/office/baseframework/view/component'
], function (Utils, Forms, Component) {

    'use strict';

    // private static functions ===============================================

    /**
     * Returns the container node of the passed control group.
     */
    function getContainerNode(group) {
        return group.getNode().parent();
    }

    /**
     * Updates the visibility of the container node of the passed control group
     * depending if it contains at least one visible group.
     */
    function updateContainerVisibility(group) {
        var contNode = getContainerNode(group);
        var visibleGroups = Forms.filterVisibleNodes(contNode.children());
        Forms.showNodes(contNode, visibleGroups.length > 0);
    }

    // class ToolBar ==========================================================

    /**
     * Represents a view component intended to be inserted into a horizontal
     * tool pane.
     *
     * @constructor
     *
     * @extends Component
     *
     * @param {BaseView} docView
     *  The document view instance containing this tool bar.
     *
     * @param {Object} [initOptions]
     *  Optional parameters. Supports all options of the Component base class.
     */
    var ToolBar = Component.extend({ constructor: function (docView, initOptions) {

        // self reference
        var self = this;

        // the current container node as target for new groups
        var containerNode = null;

        // whether to add a separator line into the next gap
        var separator = false;

        // special classes for the separator
        var separatorClasses = null;

        // base constructor ---------------------------------------------------

        Component.call(this, docView, initOptions);

        // private methods ----------------------------------------------------

        /**
         * Prepares to add a gap before the next inserted control group.
         */
        function addGap(sep) {
            containerNode = null;
            separator = sep;
        }

        /**
         * Inserts the root DOM node of the passed group into this tool bar.
         */
        function groupInserter(group) {

            // create missing group container node
            if (!containerNode) {
                containerNode = $('<div class="group-container' + (separator ? (' separator ' + separatorClasses) : '') + '">');
                self.getNode().append(containerNode);
            }

            // insert the group into the container
            containerNode.append(group.getNode());
            updateContainerVisibility(group);

            // show the container node before showing the group, for correct group:resize events
            self.listenTo(group, 'group:beforeshow', function () {
                Forms.showNodes(getContainerNode(group), true);
            });

            // hide the container node if no visible groups remain
            self.listenTo(group, 'group:hide', function () {
                updateContainerVisibility(group);
            });
        }

        // methods ------------------------------------------------------------

        /**
         * Inserts some white-space after the last group into the tool bar.
         *
         * @returns {ToolBar}
         *  A reference to this instance.
         */
        this.addGap = function () {
            addGap(false);
            return this;
        };

        /**
         * Inserts a vertical separator line after the last group into the tool
         * bar.
         *
         * @returns {ToolBar}
         *  A reference to this instance.
         */
        this.addSeparator = function (options) {
            separatorClasses = Utils.getStringOption(options, 'classes', '');
            addGap(true);
            return this;
        };

        // initialization -----------------------------------------------------

        this.getNode().addClass('toolbar').attr({ role: 'navigation' });

        // insert new groups into the current container node
        this.registerGroupInserter(groupInserter);

        // destroy all class members on destruction
        this.registerDestructor(function () {
            self = docView = initOptions = containerNode = null;
        });

    } }); // class ToolBar

    // exports ================================================================

    return ToolBar;

});
