/**
 * This work is provided under the terms of the CREATIVE COMMONS PUBLIC
 * LICENSE. This work is protected by copyright and/or other applicable
 * law. Any use of the work other than as authorized under this license
 * or copyright law is prohibited.
 *
 * http://creativecommons.org/licenses/by-nc-sa/2.5/
 *
 * Copyright (C) 2016 OX Software GmbH
 * Mail: info@open-xchange.com
 *
 * @author Miroslav Dzunic <miroslav.dzunic@open-xchange.com>
 */

define('io.ox/office/presentation/view/dialog/slidebackgrounddialog', [
    'io.ox/office/tk/utils',
    'io.ox/office/tk/keycodes',
    'io.ox/office/tk/forms',
    'io.ox/office/tk/utils/tracking',
    'io.ox/office/tk/dialog/basedialog',
    'io.ox/office/editframework/view/editcontrols',
    'io.ox/office/textframework/view/controls',
    'io.ox/office/textframework/utils/textutils',
    'gettext!io.ox/office/presentation/main',
    'less!io.ox/office/presentation/view/dialog/presentationdialogs'
], function (Utils, KeyCodes, Forms, Tracking, BaseDialog, EditControls, Controls, TextUtils, gt) {

    'use strict';

    // convenience shortcuts
    var SpinField = EditControls.SpinField;
    var NumberValidator = SpinField.NumberValidator;

    // Global constants and variables
    var slideBackgroundTitle = /*#. Change slide background from dialog menu*/ gt('Background'),
        dialogWidth = Utils.SMALL_DEVICE ? 300 : 360;

    // the regular expression for the percentage validator
    var PERCENT_RE = /^\s*0*([1-9]\d?|100)?\s*%?\s*$/;

    // constants for pointer events
    var START_POINTER_EVENTS = Utils.IOS ? 'touchstart' : 'mousedown touchstart';
    var MOVE_POINTER_EVENTS = Utils.IOS ? 'touchmove' : 'mousemove touchmove';
    var END_POINTER_EVENTS = Utils.IOS ? 'touchend' : 'mouseup touchend';

    // class PercentValidator =================================================

    /**
     * A validator for percentage fields that restricts the allowed values to
     * integer numbers and optionally % sign. Entry must be integer value
     * between 0 and 100.
     *
     * @constructor
     *
     * @extends NumberValidator
     *
     * @param {Object} [initOptions]
     *  Optional parameters. Supports all options that are supported by the
     *  base class NumberValidator.
     */
    var PercentValidator = NumberValidator.extend(function (initOptions) {
        NumberValidator.call(this, initOptions);
    }); // class PercentValidator

    // public methods ---------------------------------------------------------

    /**
     * Converts the passed value to a display string with percentage in
     * hundreds.
     *
     * @param {Number} value
     *  The percentage in number.
     *
     * @returns {String}
     *  The display string of a percentage with percent sign.
     */
    PercentValidator.prototype.valueToText = function (value) {
        var number = Utils.isFiniteNumber(value) ? Math.round(value) : 0;
        return number + ' %';
    };

    /**
     * Converts the passed display string of a percentage to number value.
     *
     * @param {String} text
     *  The display string of a percentage with % sign.
     *
     * @returns {Number|Null}
     *  The percentage number in hundreds.
     */
    PercentValidator.prototype.textToValue = function (text) {
        var matches = PERCENT_RE.exec(text);
        return (matches && matches.length) ? parseInt(matches[0], 10) : 0;
    };

    /**
     * The validation functions that controls the input into the unit field
     * corresponding to the defined regular expression.
     *
     * @returns {Boolean}
     *  Whether the input into the unit field matches the defined regular
     *  expression.
     */
    PercentValidator.prototype.validate = function (text) {
        return PERCENT_RE.test(text);
    };

    // class SlideBackgroundDialog ============================================

    /**
     * The slide background dialog.
     * Provides different types and options for slide backgrounds for Presentation document.
     *
     *
     * The changes will be applied to the current document after clicking the 'Ok' button.
     *
     * The dialog itself is shown by calling this.execute. The caller can react on successful
     * finalization of the dialog within an appropriate done-Handler, chained to the
     * this.execute function call.
     *
     * @constructor
     *
     * @extends BaseDialog
     *
     * @param {PresentationView} docView
     *  The document view containing this dialog instance.
     *
     * @param {Object} [dialogOptions]
     *  Options for filling the dialog
     *  @param {Object} [dialogOptions.color]
     *      If background is already set as type solid color.
     *  @param {String} [dialogOptions.type]
     *      Type of background.
     */
    var SlideBackgroundDialog = BaseDialog.extend(function (docView, dialogOptions) {

        var
            self = this,
            docModel = docView.getDocModel(),
            slideId = docModel.getActiveSlideId(),
            $slide = docModel.getSlideById(slideId),
            colorPicker = new Controls.ColorPicker(docView, null, { icon: 'docs-cell-fill-color', tooltip: null, title: null }),
            imagePicker = new Controls.CompactImagePicker(docView, { label: gt('Choose'), itemKey: 'slide/setbackgroundimage' }),
            transpArea, transpContainer, transpSlider, transpSliderWidth, transpSliderOffset, transpRailOverlay, transpPointer, transpSpinner,
            colorArea, imageArea, $tempPreviewSlide,
            slidingActive = false,
            startOffset = 0,
            dialogNode = null,
            applyallBtn, headerNode,
            // container node for first column (category drop-down, fields type list)
            $controlArea = null,
            modifyContainer = null,
            hideBgCheckbox = null,
            noBgRadioBtn = null,
            colorRadioBtn = null,
            imageRadioBtn = null;

        // base constructor ---------------------------------------------------

        BaseDialog.call(this, docView, {
            title: slideBackgroundTitle,
            width: dialogWidth,
            okLabel: gt('Apply')
        });

        // private methods ----------------------------------------------------

        /**
         * Spinner wrapper.
         * @param {Number} min Value
         * @param {Number} max Value
         */
        var PercentSpin = SpinField.extend({ constructor: function (min, max) {
            var spinnerValue = 0,
                spinnerNode;

            SpinField.call(this, docView, {
                width: Utils.IOS ? 75 : 60,
                min: min,
                max: max,
                validator: new PercentValidator({ min: min, max: max, percentageUnit: true }),
                keyboard: 'text', // #31765: force keyboard type to 'text', Android does not like characters in numeric text fields
                percentageUnit: true,
                smallerVersion: { hide: true }
            });

            this.setValue = function (value) {
                var attrs = $slide.data('previewAttrs');

                spinnerValue = value;
                this.setFieldValue(spinnerValue);
                // update slider
                updateSliderValue(spinnerValue);
                setAlphaPreview(attrs, spinnerValue);

                this.getInputNode().trigger('input');
            };

            this.getValue = function () {
                return spinnerValue;
            };

            this.triggerChange = function (value) {

                if (!_.isNull(value)) {
                    this.setValue(value);
                } else {
                    this.setValue(spinnerValue);
                }
            };

            /**
             * Returns the node of the spinner
             * @returns {jQuery} the spinner node.
             */
            this.node = function () {
                if (!spinnerNode) {
                    spinnerNode = $(this.getNode());
                }
                return spinnerNode;
            };
        } });

        /**
         * Helper function for enabling or disabling the specified DOM elements.
         *
         * @param {jQuery} nodes
         *  The DOM elements to be manipulated. If this object is a jQuery
         *  collection, uses all nodes it contains.
         *
         * @param {Boolean} state
         *  Whether to enable (true) or disable (false) the DOM elements.
         */
        function enableNodes(nodes, state) {
            nodes.toggleClass(Forms.DISABLED_CLASS, !state);
            if (state) {
                nodes.removeAttr('aria-disabled');
                nodes.prop('tabindex', 0);
            } else {
                nodes.attr('aria-disabled', true);
                nodes.prop('tabindex', -1);
            }
        }

        /**
         * Generates and returns HTML markup string for color slider.
         *
         * @return {String}
         */
        function generateColorSlider() {
            var markup = '<div class="slider-container" tabindex="0"><div class="bg-color-slider"><div class="slider-rail rail-overlay"></div><div class="slider-rail"></div><div class="slider-pointer"></div></div></div>';

            return markup;
        }

        /**
         * Heleper to get client x coordinate from original event of jQuery's touch event
         * @param  {jQuery Event} event
         * @return {Number} clientX value
         */
        function getTouchClientX(event) {
            var originalEvent = event.originalEvent;
            var touch = originalEvent && originalEvent.touches && originalEvent.touches[0];
            return touch && touch.clientX;
        }

        /**
         * Set alpha with corresponding attributes to slide background as preview.
         *
         * @param {Object} attrs Fill attributes to set to slide background as preview
         * @param {Number} value Value of the transparency [0, 100]
         */
        function setAlphaPreview(attrs, value) {
            var inheritFromParent = false;
            var imageLoad = false;

            if (attrs && attrs.fill && (attrs.fill.type === null || attrs.fill.type === 'none')) { // inherit fill from parent
                var parentId = docModel.getParentSlideId(slideId);
                var parentFillAttrs = docModel.getSlideAttributesByFamily(parentId, 'fill');
                if (!parentFillAttrs || parentFillAttrs.type === 'none' || parentFillAttrs.type === null) {
                    parentId = docModel.getParentSlideId(parentId);
                    parentFillAttrs = docModel.getSlideAttributesByFamily(parentId, 'fill');
                }
                attrs = { fill: _.copy(parentFillAttrs, true) };
                if (parentFillAttrs && parentFillAttrs.color && !parentFillAttrs.bitmap) {
                    colorPicker.setValue(parentFillAttrs.color);
                }
                inheritFromParent = true;
                imageLoad = true;
            }
            if (attrs && attrs.fill && (attrs.fill.type === 'solid' || attrs.fill.type === 'scheme')) {
                attrs.fill.color = attrs.fill.color || {};
                attrs.fill.color.transformations = attrs.fill.color.transformations || [];
                var alphaT = _.findWhere(attrs.fill.color.transformations, { type: 'alpha' });
                if (alphaT) {
                    alphaT.value = _.isNumber(value) ? (100 - value) * 1000 : 0;
                } else {
                    attrs.fill.color.transformations.push({ type: 'alpha', value: (100 - value) * 1000 });
                }

                docModel.getSlideStyles().updateBackground($tempPreviewSlide, attrs, slideId, null, { preview: true });
                if (inheritFromParent) {
                    docView.handleSlideBackgroundVisibility(slideId, attrs.fill);
                }
                // store values for apply
                $slide.data('previewAttrs', attrs);
            }
            if (attrs && attrs.fill && attrs.fill.type === 'bitmap' && attrs.fill.bitmap) {
                attrs.fill.bitmap.transparency = _.isNumber(value) ? value / 100 : 0;

                docModel.getSlideStyles().updateBackground($tempPreviewSlide, attrs, slideId, null, { imageLoad: imageLoad, preview: true });
                if (inheritFromParent) {
                    docView.handleSlideBackgroundVisibility(slideId, attrs.fill);
                }
                // store values for apply
                $slide.data('previewAttrs', attrs);
            }
        }

        function updateSliderValue(value) {
            transpPointer.css('left', value + '%').data('percentage', value);
            transpRailOverlay.css('width', value + '%');
        }

        function updateSliderSpinner(value) {
            updateSliderValue(value);
            transpSpinner.setValue(value);
        }

        /**
         * Set value to the slider, corresponding percentage spinner, and make preview of transparency.
         *
         * @param  {Number} value
         */
        function updateTransparency(value) {
            var normalizedValue = Utils.minMax(Utils.round(value, 1), 0, 100);
            var attrs = $slide.data('previewAttrs');

            updateSliderSpinner(normalizedValue);
            setAlphaPreview(attrs, normalizedValue);
        }

        /**
         * Initializes mouse/touch tracking to move this menu around.
         */
        function trackingStartHandler(event) {
            dialogNode = self.getBody().parents('.io-ox-office-dialog');
            startOffset = dialogNode.offset();
            event.preventDefault();
        }

        /**
         * Moves the menu while mouse/touch tracking is active.
         */
        function trackingMoveHandler(event) {
            //console.warn(startOffset.left, event.offsetX, startOffset.top, event.offsetY);
            if (!dialogNode || !dialogNode.length) {
                Utils.error('SlideBackgroundDialog.trackingMoveHandler(): dialog node not found!');
                return;
            }
            var appWindowNode = $(window);

            var marginBoundary = 5;
            var rightBoundary = appWindowNode.width() - dialogNode.width() - marginBoundary;
            var bottomBoundary = appWindowNode.height() - (Utils.SMALL_DEVICE ? self.getHeader().outerHeight() : dialogNode.height()) - marginBoundary;

            var leftValue = Utils.minMax(startOffset.left + event.offsetX, marginBoundary, rightBoundary);
            var topValue = Utils.minMax(startOffset.top + event.offsetY, marginBoundary, bottomBoundary);
            dialogNode.css({ left: leftValue, top: topValue, margin: 0 });

            // set new value for transparency slider
            transpSliderOffset = transpSlider.offset().left;
            event.preventDefault();
        }

        /**
         * Initialize controls of the dialog.
         */
        function initControls() {
            var dialogOptionsType = (_.isObject(dialogOptions) && dialogOptions.type) || null;
            var isImageRadioEnabled = dialogOptionsType === 'bitmap';
            var isColorRadioEnabled = dialogOptionsType === 'solid' || dialogOptionsType === 'scheme';
            var transpValue = 0;

            var noBgRadio = Forms.createButtonMarkup({ attributes: { class: 'no-bg-radio', role: 'radio' }, label: gt('No background') });
            var colorRadio = Forms.createButtonMarkup({ attributes: { class: 'color-radio', role: 'radio' }, label: gt('Color') });
            var imageRadio = Forms.createButtonMarkup({ attributes: { class: 'image-radio', role: 'radio' }, label: gt('Image') });
            var noBgRadioArea = '<div class="dradio">' + noBgRadio + '</div>';
            var colorRadioArea = '<div class="dradio">' + colorRadio + '</div>';
            var imageRadioArea = '<div class="dradio">' + imageRadio + '</div>';
            var transpLabel = $('<div class="transparency-label">').text(gt('Transparency'));

            var hideBgBtn = Forms.createButtonMarkup({ attributes: { class: 'hide-bg-btn', role: 'checkbox' }, label: gt('Hide background graphics') });
            var hideBgContainer = $('<div class="hidebg-container">').append(hideBgBtn);

            if (isColorRadioEnabled && dialogOptions.color) {
                colorPicker.setValue(dialogOptions.color);
            }
            colorArea = $('<div class="color-area">').append(colorPicker.getNode());
            imageArea = $('<div class="image-area">').append(imagePicker.getNode());
            transpArea = $('<div class="transparency-area">');

            transpSpinner = new PercentSpin(0, 100);
            transpArea.append(transpLabel, generateColorSlider(), transpSpinner.getNode());
            transpContainer = transpArea.children('.slider-container');
            transpSlider = transpContainer.children('.bg-color-slider');
            transpRailOverlay = transpSlider.children('.rail-overlay');
            transpPointer = transpSlider.children('.slider-pointer');

            if (Utils.SMALL_DEVICE) {
                transpPointer.css('transform', 'scale(2)');
            }

            // don't append transparency area if app is ODF
            modifyContainer = $('<div class="modify-container">').append(noBgRadioArea, colorRadioArea, colorArea, imageRadioArea, imageArea, (docModel.getApp().isODF() ? null : transpArea));

            // append all containers to control area
            $controlArea.append(modifyContainer, hideBgContainer);

            // set value to slider/spinner
            if (_.isObject(dialogOptions) && dialogOptionsType !== 'none' && dialogOptionsType !== null) {
                if (isColorRadioEnabled && dialogOptions.color && dialogOptions.color.transformations) {
                    _.each(dialogOptions.color.transformations, function (transformation) {
                        if (transformation.type === 'alpha') {
                            transpValue = 100 - transformation.value / 1000;
                        }
                    });
                }
                if (isImageRadioEnabled && dialogOptions.bitmap && dialogOptions.bitmap.transparency) {
                    transpValue = dialogOptions.bitmap.transparency * 100;
                }
            } else { // check if bitmap is inherited from layout or master slide, and enable proper buttons
                var parentId = docModel.getParentSlideId(slideId);
                var parentFillAttrs = docModel.getSlideAttributesByFamily(parentId, 'fill');
                if (!parentFillAttrs || parentFillAttrs.type === 'none' || parentFillAttrs.type === null) {
                    parentId = docModel.getParentSlideId(parentId);
                    parentFillAttrs = docModel.getSlideAttributesByFamily(parentId, 'fill');
                }
                if (parentFillAttrs && parentFillAttrs.type === 'bitmap') {
                    isImageRadioEnabled = true;
                    transpValue = (parentFillAttrs.bitmap.transparency || 0) * 100;
                } else if (parentFillAttrs && (parentFillAttrs.type === 'solid' || parentFillAttrs.type === 'scheme')) {
                    isColorRadioEnabled = true;
                    var alphaT = (parentFillAttrs.color && parentFillAttrs.color.transformations) ? _.findWhere(parentFillAttrs.color.transformations, { type: 'alpha' }) : 0;
                    transpValue = (alphaT && _.isNumber(alphaT.value) && (100 - alphaT.value / 1000)) || 0;
                }
            }
            transpSpinner.setValue(transpValue);

            // enable/disable nodes
            enableNodes(colorArea.find('a'), isColorRadioEnabled);
            enableNodes(imageArea.find('a'), isImageRadioEnabled);
            enableNodes(transpArea.find('.slider-container, input'), isImageRadioEnabled || isColorRadioEnabled);
            imageArea.prop('disabled', !isImageRadioEnabled).toggleClass('disabled', !isImageRadioEnabled);
            colorArea.prop('disabled', !isColorRadioEnabled).toggleClass('disabled', !isColorRadioEnabled);
            transpArea.toggleClass('disabled', !isImageRadioEnabled && !isColorRadioEnabled);
            enableNodes(hideBgContainer.find('.hide-bg-btn'), !docModel.isMasterSlideId(docModel.getActiveSlideId()));

            // no-background radio
            noBgRadioBtn = $controlArea.find('.no-bg-radio');
            Forms.checkButtonNodes(noBgRadioBtn, !isImageRadioEnabled && !isColorRadioEnabled, { design: 'radio', ambiguous: false });
            Forms.setButtonKeyHandler(noBgRadioBtn);
            // color radio
            colorRadioBtn = $controlArea.find('.color-radio');
            Forms.checkButtonNodes(colorRadioBtn, isColorRadioEnabled, { design: 'radio', ambiguous: false });
            Forms.setButtonKeyHandler(colorRadioBtn);
            // image radio
            imageRadioBtn = $controlArea.find('.image-radio');
            Forms.checkButtonNodes(imageRadioBtn, isImageRadioEnabled, { design: 'radio', ambiguous: false });
            Forms.setButtonKeyHandler(imageRadioBtn);

            // hide background checkbox
            hideBgCheckbox = $controlArea.find('.hide-bg-btn');
            Forms.checkButtonNodes(hideBgCheckbox, !docModel.isFollowMasterShapeSlide(), { design: 'boxed', ambiguous: false });
            Forms.setButtonKeyHandler(hideBgCheckbox);
        }

        // initialization -----------------------------------------------------

        // store deep cloned original values in node data
        $slide.data('previewAttrs', { fill: dialogOptions ? _.copy(dialogOptions, true) : TextUtils.getEmptySlideBackgroundAttributes().fill, slide: { followMasterShapes: docModel.isFollowMasterShapeSlide() } });

        // close dialog when losing edit rights
        docView.closeDialogOnReadOnlyMode(this);

        $controlArea = $('<div class="control-area">');
        $tempPreviewSlide = docModel.getHelpBackgroundNode($slide);

        // initialize the body element of the dialog
        this.getBody()
            .addClass('io-ox-office-presentation-dialog-slide-background')
            .toggleClass('mobile', !!Utils.SMALL_DEVICE)
            .append($controlArea);

        // create the layout of the dialog
        initControls();

        if (dialogOptions && dialogOptions.type !== 'none' && dialogOptions.type !== null) {
            docModel.getSlideStyles().updateBackground($tempPreviewSlide, { fill: dialogOptions || { type: null } }, slideId, null, { preview: true, imageLoad: true, onDialogOpen: true });
        }

        // add additional button
        applyallBtn = this.createButton('applyall', gt('Apply to all'), { buttonStyle: 'primary' });

        // ensure proper button order on small devices, because base dialog makes reorder
        if (Utils.SMALL_DEVICE) {
            this.on('beforeshow', function () {
                applyallBtn.parent().detach();
                this.getFooter().children('.row').append(applyallBtn.parent());
            });
        } else {
            this.getFooter().find('button').addClass('truncate-footer-btn');
        }

        // set initial control values
        this.on('show', function () {
            // blocking keyboard input during applying of operations
            docModel.setBlockKeyboardEvent(true);

            transpSliderWidth = transpSlider.width();
            transpSliderOffset = transpSlider.offset().left;
        });

        // handler for the OK button
        this.setOkHandler(function () {
            var startOptions = _.extend({ fill: dialogOptions }, { slide: { followMasterShapes: docModel.isFollowMasterShapeSlide() } });
            var endOptions = $slide.data('previewAttrs') || {};
            endOptions = _.extend({ slide: { followMasterShapes: docModel.isFollowMasterShapeSlide() } }, endOptions);
            return _.isEqual(startOptions, endOptions) ?
                docModel.removeHelpBackgroundNode($slide) :
                docModel.applySlideBackgroundOperations($slide);
        });

        // handler for the "Apply All" button
        this.setActionHandler('applyall', function () {
            docModel.getSlideStyles().updateBackground($tempPreviewSlide, { fill: { type: null } }, slideId, null, { preview: true });
            docModel.removeHelpBackgroundNode($slide);
            return docModel.applySlideBackgroundOperations($slide, { allSlides: true });
        });

        // on cancel restore original values
        this.setCancelHandler(function () {
            var attrs = dialogOptions || { type: null };
            docModel.getSlideStyles().updateBackground($tempPreviewSlide, { fill: attrs }, slideId, null, { preview: true });
            docView.previewFollowMasterPageChange({ followMasterShapes: docModel.isFollowMasterShapeSlide() });
            docModel.removeHelpBackgroundNode($slide);
            docView.handleSlideBackgroundVisibility(slideId);
        });

        // listen on checkbox to change icon
        hideBgCheckbox.on('click', function () {
            var slideData = $slide.data('previewAttrs');
            Forms.checkButtonNodes(this, !Forms.isCheckedButtonNode(this), { design: 'boxed', ambiguous: false });
            $slide.data('previewAttrs', _.extend(slideData, { slide: { followMasterShapes: !Forms.isCheckedButtonNode(this) } }));
            docView.previewFollowMasterPageChange({ followMasterShapes: !Forms.isCheckedButtonNode(this) });
        });

        noBgRadioBtn.on('click', function () {
            Forms.checkButtonNodes(colorRadioBtn.add(imageRadioBtn), false, { design: 'radio', ambiguous: false });
            Forms.checkButtonNodes(this, true, { design: 'radio', ambiguous: false });
            colorArea.prop('disabled', true).addClass('disabled');
            imageArea.prop('disabled', true).addClass('disabled');
            transpArea.addClass('disabled');
            enableNodes(transpArea.find('.slider-container, input'), false);
            enableNodes(colorArea.find('a'), false);
            enableNodes(imageArea.find('a'), false);

            var attrs = TextUtils.getEmptySlideBackgroundAttributes();
            transpSpinner.setValue(0);
            colorPicker.setValue({ type: 'auto' });
            docView.handleSlideBackgroundVisibility(slideId, attrs.fill);
            docModel.getSlideStyles().updateBackground($tempPreviewSlide, attrs, slideId, null, { preview: true });
            $slide.data('previewAttrs', attrs);
        });

        colorRadioBtn.on('click', function () {
            Forms.checkButtonNodes(imageRadioBtn.add(noBgRadioBtn), false, { design: 'radio', ambiguous: false });
            Forms.checkButtonNodes(this, true, { design: 'radio', ambiguous: false });
            colorArea.prop('disabled', false).removeClass('disabled');
            imageArea.prop('disabled', true).addClass('disabled');
            transpArea.removeClass('disabled');
            enableNodes(transpArea.find('.slider-container, input'), true);
            enableNodes(colorArea.find('a'), true);
            enableNodes(imageArea.find('a'), false);
        });

        imageRadioBtn.on('click', function () {
            Forms.checkButtonNodes(colorRadioBtn.add(noBgRadioBtn), false, { design: 'radio', ambiguous: false });
            Forms.checkButtonNodes(this, true, { design: 'radio', ambiguous: false });
            imageArea.prop('disabled', false).removeClass('disabled');
            colorArea.prop('disabled', true).addClass('disabled');
            transpArea.removeClass('disabled');
            enableNodes(transpArea.find('.slider-container, input'), true);
            enableNodes(imageArea.find('a'), true);
            enableNodes(colorArea.find('a'), false);
        });

        // transparency slider events handling
        transpSlider.on(START_POINTER_EVENTS, function (event) {
            var clientX = event.type === 'mousedown' ? event.clientX : getTouchClientX(event);
            var leftPos = (clientX - transpSliderOffset) * 100 / transpSliderWidth;
            slidingActive = true;
            updateTransparency(leftPos);

            docModel.listenTo(self.getBody(), MOVE_POINTER_EVENTS, function (event) {
                var clientX = event.type === 'mousemove' ? event.clientX : getTouchClientX(event);
                var leftPos = (clientX - transpSliderOffset) * 100 / transpSliderWidth;
                updateTransparency(leftPos);
            });
        });

        this.getBody().on(END_POINTER_EVENTS, function () {
            if (slidingActive) {
                docModel.stopListeningTo(self.getBody(), MOVE_POINTER_EVENTS);
                slidingActive = false;
            }
        });

        transpContainer.on('keydown', function (event) {
            var perc = transpPointer.data('percentage') || 0;
            if (event.keyCode === KeyCodes.UP_ARROW || event.keyCode === KeyCodes.RIGHT_ARROW) {
                updateTransparency(perc + 1);
            }
            if (event.keyCode === KeyCodes.DOWN_ARROW || event.keyCode === KeyCodes.LEFT_ARROW) {
                updateTransparency(perc - 1);
            }
        });

        // color picker event handling
        colorPicker.on('group:change', function (event, data) {
            var attrs = data.type === 'auto' ? TextUtils.getEmptySlideBackgroundAttributes() : { fill: { type: 'solid', color: data, imageUrl: null } };

            docView.handleSlideBackgroundVisibility(slideId, attrs.fill);
            setAlphaPreview(attrs, transpSpinner.getValue());
            this.hideMenu();
        });

        headerNode = this.getHeader().css('cursor', 'move');
        Tracking.enableTracking(headerNode);
        headerNode.on({ 'tracking:start': trackingStartHandler, 'tracking:move': trackingMoveHandler });

        this.registerDestructor(function () {
            // blocking keyboard input during applying of operations
            docModel.setBlockKeyboardEvent(false);
            $slide.removeData('previewAttrs');
            if (colorPicker) { colorPicker.destroy(); }
            if (imagePicker) { imagePicker.destroy(); }
            if (transpSpinner) { transpSpinner.destroy(); }

            $controlArea = modifyContainer = hideBgCheckbox = noBgRadioBtn = colorRadioBtn = imageRadioBtn = null;
            self = docModel = slideId = $slide = colorPicker = imagePicker = colorArea = imageArea = null;
            transpArea = transpContainer = transpSlider = transpSliderWidth = transpSliderOffset = transpRailOverlay = transpPointer = null;
            transpSpinner = $tempPreviewSlide = applyallBtn = headerNode = null;
        });

    }); // class SlideBackgroundDialog

    // exports ================================================================

    return SlideBackgroundDialog;

});
